"""Define S3Uploader class."""
import os

from boto3 import client
from ML_management.uploader_data.utils import get_space_size, get_upload_paths
from tqdm import tqdm


class S3Uploader:
    """S3 uploader files class."""

    def __init__(self):
        """Init creds."""
        self.default_url = os.environ.get("MLFLOW_S3_ENDPOINT_URL", "http://minio-mlmanagement.ru/")
        self.default_access_key_id = os.environ.get("AWS_ACCESS_KEY_ID", "minioadmin")
        self.default_secret_access_key = os.environ.get("AWS_SECRET_ACCESS_KEY", "minioadmin")

    def upload(self, local_path: str, bucket: str, verbose: bool = True):
        """Upload."""
        local_path = os.path.normpath(local_path)
        if not os.path.exists(local_path):
            raise FileExistsError(f"Path: {local_path} does not exist")

        service_client = client(
            service_name="s3",
            use_ssl=True,
            verify=None,
            endpoint_url=self.default_url,
            aws_access_key_id=self.default_access_key_id,
            aws_secret_access_key=self.default_secret_access_key,
        )
        buckets = [item["Name"] for item in service_client.list_buckets()["Buckets"]]
        if bucket not in buckets:
            service_client.create_bucket(Bucket=bucket)

        space_size = get_space_size(local_path)
        upload_paths = get_upload_paths(local_path)

        with tqdm(
            total=space_size,
            disable=not verbose,
            unit="B",
            unit_scale=True,
            unit_divisor=1024,
        ) as pbar:
            for path in upload_paths:
                with open(path.local_path, "rb") as _file:
                    service_client.upload_fileobj(
                        Fileobj=_file,
                        Bucket=bucket,
                        Key=path.storage_path,
                        Callback=pbar.update,
                    )



"""Define Abstract class for Model with necessary methods and methods to implement."""
from abc import ABC, abstractmethod
from typing import List, Optional

from ML_management.mlmanagement import mlmanagement, variables
from ML_management.mlmanagement.module_finder import ModuleFinder

import mlflow


class Model(mlflow.pyfunc.PythonModel, ABC):
    """Abstract class for model that Job will use."""

    def __init__(self):
        """Initialize Model class."""
        # That parameters will be set automatically while loading the model.
        """
        :param self.artifacts: A dictionary containing ``<name, local_path>`` entries.
        """
        self.artifacts: dict = {}
        # That parameters will be set automatically in job before the 'execute' func would be executed.
        """
        "param self.dataset: DatasetLoader object
        """
        self.dataset = None

    @abstractmethod
    def predict_function(self, **kwargs):
        """Every model should make predictions."""
        raise NotImplementedError

    def upload_model(
        self,
        registered_model_name: str,
        artifacts: dict = None,
        experiment_name: str = None,
        pip_requirements=None,
        extra_pip_requirements=None,
        conda_env=None,
        extra_modules_names: Optional[List[str]] = None,
        used_modules_names: Optional[List[str]] = None,
    ):
        """
        Upload wrapper to MLmanagement server.

        :param pyfunc_model_name: The run-relative artifact path to which to log the Python model.
        :param artifacts: A dictionary containing ``<name, artifact_uri>`` entries. Remote artifact URIs
                          are resolved to absolute filesystem paths, producing a dictionary of
                          ``<name, absolute_path>`` entries. ``python_model`` can reference these
                          resolved entries as the ``artifacts`` property of the ``context`` parameter
                          in :func:`PythonModel.load_context() <mlflow.pyfunc.PythonModel.load_context>`
                          and :func:`PythonModel.predict() <mlflow.pyfunc.PythonModel.predict>`.
                          For example, consider the following ``artifacts`` dictionary::

                            {
                                "my_file": "s3://my-bucket/path/to/my/file"
                            }

                          In this case, the ``"my_file"`` artifact is downloaded from S3. The
                          ``python_model`` can then refer to ``"my_file"`` as an absolute filesystem
                          path via ``context.artifacts["my_file"]``.

                          If ``None``, no artifacts are added to the model.
        :param registered_model_name: This argument may change or be removed in a
                                      future release without warning. If given, create a model
                                      version under ``registered_model_name``, also creating a
                                      registered model if one with the given name does not exist
        :param experiment_name: Name of experiment to which load the model
        :param pip_requirements: {{ pip_requirements }}
        :param extra_pip_requirements: {{ extra_pip_requirements }}
        `pip_requirements` and 'extra_pip_requirements' must be either a string path to a pip requirements file on the
            local filesystem or an iterable of pip requirement strings.
        :param conda_env: {{ conda_env }}
        'conda_env' must be a dict specifying the conda environment for this model.
        :param extra_modules_names: names of modules that should be pickled by value
            in addition to auto-detected modules.
        :param used_modules_names: modules that should be pickled by value, disables the auto-detection of modules.
        """
        old_experiment_name = variables.active_experiment_name
        if experiment_name:
            mlmanagement.set_experiment(experiment_name)
        try:
            with mlmanagement.start_run(nested=True):
                mlmanagement.log_model(
                    artifact_path="",
                    python_model=self,
                    artifacts=artifacts,
                    registered_model_name=self.model_name if hasattr(self, "model_name") else registered_model_name,
                    source_model_name=self.source_model_name
                    if hasattr(self, "source_model_name")
                    else None,  # set that after model download from mlflow
                    source_model_version=self.source_model_version
                    if hasattr(self, "source_model_version")
                    else None,  # set that after model download from mlflow
                    pip_requirements=pip_requirements,
                    extra_pip_requirements=extra_pip_requirements,
                    conda_env=conda_env,
                    extra_modules_names=extra_modules_names,
                    used_modules_names=used_modules_names,
                    root_module_name=ModuleFinder.get_my_caller_module_name(),
                )
        except Exception as err:
            raise err
        finally:
            variables.active_experiment_name = old_experiment_name



"""
Mlflow server inside ML_management.

ML_manager sends request to our server on /mlflow endpoint
Server calls original mlflow function accordingly
"""
import atexit
import inspect
import io
import os
import subprocess
import sys
import tempfile
import traceback
import zipfile
from typing import Any, Dict, List, Optional, Union

import cloudpickle
import numpy
import pandas
from ML_management.mlmanagement import variables
from ML_management.mlmanagement.abstract_mlflow_client import AbstractMlflowClient
from ML_management.mlmanagement.mlmanager import request_for_function
from ML_management.mlmanagement.model_type import ModelType
from ML_management.mlmanagement.module_finder import ModuleFinder
from ML_management.mlmanagement.server_mlmanager_exceptions import ModelTypeIsNotFound
from ML_management.mlmanagement.utils import _load_model_type
from ML_management.mlmanagement.variables import active_run_stack
from mlflow.entities import Experiment, Run, RunStatus
from mlflow.entities.model_registry import ModelVersion, RegisteredModel
from mlflow.models import Model
from mlflow.store.entities import PagedList
from mlflow.tracking.fluent import _RUN_ID_ENV_VAR
from mlflow.utils import env
from scipy.sparse import csc_matrix, csr_matrix

import mlflow
from mlflow import ActiveRun


def monkey_patching_exit(self, exc_type, exc_val, exc_tb):
    """Redefine __exit__ function of class ActiveRun."""
    status = RunStatus.FINISHED if exc_type is None else RunStatus.FAILED
    end_run(RunStatus.to_string(status))
    return exc_type is None


# Rewrite __exit__ method to enable using Python ``with`` syntax of ActiveRun class.
ActiveRun.__exit__ = monkey_patching_exit


def start_run_if_not_exist():
    """If run doesn't exist call start_run() function."""
    if len(active_run_stack) == 0:
        start_run()


def set_experiment(experiment_name: Optional[str] = None, experiment_id: Optional[str] = None) -> Experiment:
    """
    Set the given experiment as the active experiment.

    The experiment must either be specified by name via
    experiment_name or by ID via experiment_id. The experiment name and ID cannot both be specified.
    Set global variable active_experiment_name to that experiment_name.
    """
    request_for_function(inspect.currentframe())
    variables.active_experiment_name = experiment_name


def get_experiment_by_name(name: str) -> Optional[Experiment]:
    """Retrieve an experiment by experiment name from the backend store."""
    return request_for_function(inspect.currentframe())


def search_runs(
    experiment_ids: Optional[List[str]] = None,
    filter_string: str = "",
    run_view_type: int = 1,
    max_results: int = 100000,
    order_by: Optional[List[str]] = None,
    output_format: str = "pandas",
    search_all_experiments: bool = False,
    experiment_names: Optional[List[str]] = None,
) -> Union[List[Run], "pandas.DataFrame"]:
    """Search experiments that fit the search criteria."""
    return request_for_function(inspect.currentframe())


def start_run(
    run_id: Optional[str] = None,
    experiment_id: Optional[str] = None,
    run_name: Optional[str] = None,
    nested: bool = False,
    tags: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
) -> ActiveRun:
    """
    Start a new MLflow run, setting it as the active run under which metrics and parameters will be logged.

    The return value can be used as a context manager within a with block; otherwise, you must call end_run() to
    terminate the current run.
    If you pass a run_id or the MLFLOW_RUN_ID environment variable is set, start_run attempts to resume a run with
    the specified run ID and other parameters are ignored. run_id takes precedence over MLFLOW_RUN_ID.
    If resuming an existing run, the run status is set to RunStatus.RUNNING.
    Add that created run to active_run_stack.
    """
    if len(active_run_stack) > 0 and not nested:
        raise Exception(
            (
                "Run with UUID {} is already active. To start a new run, first end the "
                + "current run with mlmanagement.end_run(). To start a nested "
                + "run, call start_run with nested=True"
            ).format(active_run_stack[0].info.run_id)
        )
    _active_run = request_for_function(inspect.currentframe())
    active_run_stack.append(_active_run)
    return _active_run


def _add_job_registration_metainfo(
    registration_timestamp,
    init_model_version_run_id,
    dataset_loader_version_run_id,
    executor_version_run_id,
    name,
    job_executor_name,
    job_executor_version,
    dataset_loader_name,
    dataset_loader_version,
    model_name,
    model_version,
    new_model_name,
    experiment_name,
    periodic_type,
    cron_expression,
    metric_name,
    optimal_min,
    choice_criteria,
    executor_model_params_json,
    data_params_json,
    collector_name,
    gpu,
):
    """Add job metainfo on job registration. For internal use only."""
    return request_for_function(inspect.currentframe())


def _add_job_start_metainfo(
    job_name,
    start_timestamp,
):
    """Add job metainfo on execution start. For internal use only."""
    return request_for_function(inspect.currentframe())


def _add_job_end_metainfo(
    job_name,
    end_timestamp,
):
    """Add job metainfo on execution end. For internal use only."""
    return request_for_function(inspect.currentframe())


def _add_job_fail_metainfo(
    job_name,
    end_timestamp,
):
    """Add job metainfo on execution fail. For internal use only."""
    return request_for_function(inspect.currentframe())


def log_model(
    artifact_path,
    loader_module=None,
    data_path=None,
    code_path=None,
    conda_env=None,
    python_model=None,
    artifacts=None,
    registered_model_name="default_name",
    signature: mlflow.models.signature.ModelSignature = None,
    input_example: Union[pandas.core.frame.DataFrame, numpy.ndarray, dict, list, csr_matrix, csc_matrix, str, bytes] = None,
    await_registration_for=300,
    pip_requirements=None,
    extra_pip_requirements=None,
    metadata=None,
    source_model_name=None,
    source_model_version=None,
    upload_model_mode=None,
    extra_modules_names=None,
    used_modules_names=None,
    root_module_name: str = "__main__",
):
    """
    Log a Pyfunc model with custom inference logic and optional data dependencies as an MLflow artifact.

    Current run is using.
    You cannot specify the parameters: loader_module, data_path and the parameters: python_model, artifacts together.
    """
    if extra_modules_names and used_modules_names:
        raise RuntimeError("Parameters 'extra_modules_names' and 'used_modules_names' cannot be used at the same time.")
    start_run_if_not_exist()
    if used_modules_names is not None:
        submodules = set(used_modules_names)
        ModuleFinder.import_modules_by_name(submodules)
    else:
        submodules = ModuleFinder().find_root_submodules(root_name=root_module_name)
        if extra_modules_names:
            extra_set = set(extra_modules_names)
            ModuleFinder.import_modules_by_name(extra_set)
            submodules = submodules.union(extra_set)
    try:
        for module_name in submodules:
            cloudpickle.register_pickle_by_value(sys.modules[module_name])
        result = request_for_function(inspect.currentframe(), ["pyfunc"])
        return result
    finally:
        for module_name in submodules:
            cloudpickle.unregister_pickle_by_value(sys.modules[module_name])


def log_metric(key: str, value: float, step: Optional[int] = None) -> None:
    """Log a metric under the current run. If no run is active, this method will create a new active run."""
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe())


def log_artifact(local_path: str, artifact_path: Optional[str] = None) -> None:
    """
    Log a local file or directory as an artifact of the currently active run.

    If no run is active, this method will create a new active run.
    """
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe())


def log_param(key: str, value: Any) -> Any:
    """
    Log a parameter (e.g. model hyperparameter) under the current run.

    If no run is active, this method will create a new active run.
    """
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe())


def log_params(params: Dict[str, Any]) -> None:
    """
    Log a batch of params for the current run.

    If no run is active, this method will create a new active run.
    """
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe())


def set_tag(key: str, value: Any) -> None:
    """Set a tag under the current run. If no run is active, this method will create a new active run."""
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe())


def autolog(
    log_every_n_epoch=1,
    log_every_n_step=None,
    log_models=True,
    log_datasets=True,
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    registered_model_name=None,
) -> None:
    """
    Enable (or disable) and configure autologging for all supported integrations.

    The parameters are passed to any autologging integrations that support them.
    """
    start_run_if_not_exist()
    return request_for_function(inspect.currentframe(), ["pytorch"])


default_model = Model()


def save_model(
    path, loader_module=None, data_path=None, code_path=None, conda_env=None, mlflow_model=default_model, python_model=None, artifacts=None
):
    """
    Save a Pyfunc model with custom inference logic and optional data dependencies to a path on the local filesystem.

    You cannot specify the parameters: loader_module, data_path and the parameters: python_model, artifacts together.
    """
    return request_for_function(inspect.currentframe(), ["pyfunc"])


def active_run() -> Optional[ActiveRun]:
    """Get the currently active Run, or None if no such run exists."""
    return active_run_stack[-1] if len(active_run_stack) > 0 else None


def get_run(run_id: str) -> Run:
    """
    Fetch the run from backend store.

    The resulting Run contains a collection of run metadata – RunInfo, as well as a collection of run parameters,
    tags, and metrics – RunData.
    In the case where multiple metrics with the same key are logged for the run, the RunData contains the most recently
    logged value at the largest step for each metric.
    """
    return request_for_function(inspect.currentframe())


finished_run_status = RunStatus.to_string(RunStatus.FINISHED)


def end_run(status: str = finished_run_status) -> None:
    """End an active MLflow run (if there is one)."""
    if len(active_run_stack) > 0:
        # Clear out the global existing run environment variable as well.
        env.unset_variable(_RUN_ID_ENV_VAR)
        run = active_run_stack[-1]
        MlflowClient().set_terminated(run.info.run_id, status)
        active_run_stack.pop()


def set_model_version_requirements(name: str, version: int, model_type: ModelType) -> None:
    """Installing requirements of the model locally."""
    if model_type == ModelType.MODEL:
        requirements = MlflowClient().get_model_version_requirements(name=name, version=version)
    elif model_type == ModelType.EXECUTOR:
        requirements = MlflowClient().get_executor_version_requirements(name=name, version=version)
    elif model_type == ModelType.DATASET_LOADER:
        requirements = MlflowClient().get_dataset_loader_version_requirements(name=name, version=version)
    else:
        raise ModelTypeIsNotFound()
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", "--no-cache-dir", "--default-timeout=100", *requirements])
    except Exception:
        print(traceback.format_exc())
        print(f"Requirements for {name} weren't installed correctly.")


def load_model(name: str, version: int, install_requirements: bool = False, unwrap: bool = True):
    """Download all model's files for loading model locally."""
    if install_requirements:
        set_model_version_requirements(name=name, version=version, model_type=ModelType.MODEL)
    run_id = MlflowClient().get_model_version(name, version).run_id
    return _load_model_type(run_id, unwrap)


def load_dataset(name: str, version: int, install_requirements: bool = False, unwrap: bool = True):
    """Download all model's files for loading model locally."""
    if install_requirements:
        set_model_version_requirements(name=name, version=version, model_type=ModelType.DATASET_LOADER)
    run_id = MlflowClient().get_dataset_loader_version(name, version).run_id
    return _load_model_type(run_id, unwrap)


def load_executor(name: str, version: int, install_requirements: bool = False, unwrap: bool = True):
    """Download all model's files for loading model locally."""
    if install_requirements:
        set_model_version_requirements(name=name, version=version, model_type=ModelType.EXECUTOR)
    run_id = MlflowClient().get_executor_version(name, version).run_id
    return _load_model_type(run_id, unwrap)


atexit.register(end_run)


class MlflowClient(AbstractMlflowClient):
    """Initialize an MLflow Client."""

    def __init__(self):
        self.extra_attrs = ["tracking"]
        self.for_class = {"class_name": self.__class__.__name__, "class_kwargs": {}}

    def set_model_version_tag(
        self, name: str, version: Optional[str] = None, key: Optional[str] = None, value: Optional[Any] = None, stage: Optional[str] = None
    ) -> None:
        """Set model version tag."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def set_dataset_loader_version_tag(
        self, name: str, version: Optional[str] = None, key: Optional[str] = None, value: Optional[Any] = None, stage: Optional[str] = None
    ) -> None:
        """Set dataset loader version tag."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def set_executor_version_tag(
        self, name: str, version: Optional[str] = None, key: Optional[str] = None, value: Optional[Any] = None, stage: Optional[str] = None
    ) -> None:
        """Set executor version tag."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_run(self, run_id: str) -> Run:
        """Get run by id."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_registered_model(self, name: str) -> RegisteredModel:
        """Get registered model by name."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_registered_dataset_loader(self, name: str) -> RegisteredModel:
        """Get registered dataset loader by name."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_registered_executor(self, name: str) -> RegisteredModel:
        """Get registered executor by name."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    # TODO after updating to a new version, the result is paginated
    # and needs to be converted to pagination scheme of our project.
    def search_model_versions(
        self,
        filter_string: Optional[str] = None,
        max_results: int = 10000,
        order_by: Optional[List[str]] = None,
        page_token: Optional[str] = None,
    ) -> PagedList[ModelVersion]:
        """Search for model versions in backend that satisfy the filter criteria."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def set_terminated(self, run_id: str, status: Optional[str] = None, end_time: Optional[int] = None) -> None:
        """Set a run’s status to terminated."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_model_version(self, name: str, version: int) -> ModelVersion:
        """Get model version by model name and version number."""
        return request_for_function(
            inspect.currentframe(),
            for_class=self.for_class,
        )

    def get_dataset_loader_version(self, name: str, version: int) -> ModelVersion:
        """Get dataset loader version by dataset loader name and version number."""
        return request_for_function(
            inspect.currentframe(),
            for_class=self.for_class,
        )

    def get_executor_version(self, name: str, version: int) -> ModelVersion:
        """Get executor version by executor name and version number."""
        return request_for_function(
            inspect.currentframe(),
            for_class=self.for_class,
        )

    def download_artifacts(self, run_id: str, path: str, dst_path: Optional[str] = None) -> str:
        """Download an artifact file or directory from a run to a local directory, and return a local path for it."""
        file = request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

        # save artifacts to local store
        if dst_path is None:
            dst_path = tempfile.mkdtemp()
        dst_path = os.path.abspath(os.path.normpath(dst_path))
        local_path = os.path.join(dst_path, os.path.normpath(path))
        if zipfile.is_zipfile(io.BytesIO(file)):
            zip_model = zipfile.ZipFile(io.BytesIO(file))
            try:
                if not os.path.exists(local_path):
                    os.makedirs(local_path)
            except OSError as ex:
                if ex.errno == 2:
                    ex.filename = dst_path
                raise ex
            zip_model.extractall(local_path)
        else:
            dirs = os.path.dirname(local_path)
            if not os.path.exists(dirs):
                os.makedirs(dirs)
            with open(local_path, "wb") as f:
                f.write(file)
        return local_path

    def get_model_version_requirements(self, name: str, version: int) -> list:
        """Download an artifact file and return list of requirements of the model."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_model_version_conda_env(self, name: str, version: int) -> dict:
        """Download an artifact file and return json file of conda environment."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_dataset_loader_version_requirements(self, name: str, version: int) -> list:
        """Download an artifact file and return list of requirements of the model."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_dataset_loader_version_conda_env(self, name: str, version: int) -> dict:
        """Download an artifact file and return json file of conda environment."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_executor_version_requirements(self, name: str, version: int) -> list:
        """Download an artifact file and return list of requirements of the model."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )

    def get_executor_version_conda_env(self, name: str, version: int) -> dict:
        """Download an artifact file and return json file of conda environment."""
        return request_for_function(
            inspect.currentframe(),
            extra_attrs=self.extra_attrs,
            for_class=self.for_class,
        )



"""Base classes for custom exceptions used in project."""


class MLMBaseException(Exception):
    """Base exception for all custom exceptions."""

    pass


class ServerException(MLMBaseException):
    """Base exception for all server mlmanager exceptions."""

    pass


class RegistyException(MLMBaseException):
    """Base exception for all registry exceptions."""

    pass



"""Define abstract mlflow client."""

from abc import abstractmethod

from mlflow.entities import Run
from mlflow.entities.model_registry import ModelVersion, RegisteredModel
from mlflow.store.entities import PagedList


class AbstractMlflowClient:
    """Initialize an MLflow Client."""

    @abstractmethod
    def set_model_version_tag(self, *args, **kwargs) -> None:
        """Set model version tag."""
        raise NotImplementedError

    @abstractmethod
    def set_dataset_loader_version_tag(self, *args, **kwargs) -> None:
        """Set dataset loader version tag."""
        raise NotImplementedError

    @abstractmethod
    def set_executor_version_tag(self, *args, **kwargs) -> None:
        """Set executor version tag."""
        raise NotImplementedError

    @abstractmethod
    def get_run(self, *args, **kwargs) -> Run:
        """Get run by id."""
        raise NotImplementedError

    @abstractmethod
    def get_registered_model(self, *args, **kwargs) -> RegisteredModel:
        """Get registered model by name."""
        raise NotImplementedError

    @abstractmethod
    def get_registered_dataset_loader(self, *args, **kwargs) -> RegisteredModel:
        """Get registered dataset loader by name."""
        raise NotImplementedError

    @abstractmethod
    def get_registered_executor(self, *args, **kwargs) -> RegisteredModel:
        """Get registered executor by name."""
        raise NotImplementedError

    @abstractmethod
    def search_model_versions(self, *args, **kwargs) -> PagedList[ModelVersion]:
        """Search for model versions in backend that satisfy the filter criteria."""
        raise NotImplementedError

    @abstractmethod
    def set_terminated(self, *args, **kwargs) -> None:
        """Set a run’s status to terminated."""
        raise NotImplementedError

    @abstractmethod
    def get_model_version(self, *args, **kwargs) -> ModelVersion:
        """Get model version by model name and version number."""
        raise NotImplementedError

    @abstractmethod
    def get_dataset_loader_version(self, *args, **kwargs) -> ModelVersion:
        """Get model version by dataset loader name and version number."""
        raise NotImplementedError

    @abstractmethod
    def get_executor_version(self, *args, **kwargs) -> ModelVersion:
        """Get model version by executor name and version number."""
        raise NotImplementedError

    @abstractmethod
    def download_artifacts(self, *args, **kwargs) -> str:
        """Download an artifact file or directory from a run to a local directory, and return a local path for it."""
        raise NotImplementedError

    @abstractmethod
    def get_model_version_requirements(self, *args, **kwargs) -> list:
        """Get requirements of the model."""
        raise NotImplementedError

    @abstractmethod
    def get_model_version_conda_env(self, *args, **kwargs) -> dict:
        """Get conda.yaml file of the model in dictionary format."""
        raise NotImplementedError

    @abstractmethod
    def get_dataset_loader_version_requirements(self, *args, **kwargs) -> list:
        """Get requirements of the dataset loader."""
        raise NotImplementedError

    @abstractmethod
    def get_dataset_loader_version_conda_env(self, *args, **kwargs) -> dict:
        """Get conda.yaml file of the dataset loader in dictionary format."""
        raise NotImplementedError

    @abstractmethod
    def get_executor_version_requirements(self, *args, **kwargs) -> list:
        """Get requirements of the executor."""
        raise NotImplementedError

    @abstractmethod
    def get_executor_version_conda_env(self, *args, **kwargs) -> dict:
        """Get conda.yaml file of the executor in dictionary format."""
        raise NotImplementedError



"""Custom exception definition for server mlflow manager graph."""

from ML_management.mlmanagement.base_exceptions import ServerException


class InvalidEnumType(ServerException):
    """Define InvalidEnumType exception."""

    def __init__(self, passed_enum_values, enum_type_name):
        self.passed_enum_values = passed_enum_values
        self.enum_type_name = enum_type_name
        self.message = f'Passed enum values "{", ".join(passed_enum_values)}" is invalid, ' f"must be value of Enum {enum_type_name}."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidEnumType, (self.passed_enum_values, self.enum_type_name))


class InvalidModelName(ServerException):
    """Define InvalidModelName exception."""

    def __init__(self, passed_name, kwarg):
        self.passed_name = passed_name
        self.kwarg = kwarg
        self.message = 'Passed "{kwarg}" argument value "{passed_name}" is invalid, as this model is not registered in mlflow.'.format(
            kwarg=kwarg,
            passed_name=passed_name,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidModelName, (self.passed_name, self.kwarg))


class InvalidModelVersion(ServerException):
    """Define InvalidModelVersion exception."""

    def __init__(self, passed_name, passed_version, name_kwarg, version_kwarg):
        self.passed_version = passed_version
        self.passed_name = passed_name
        self.name_kwarg = name_kwarg
        self.version_kwarg = version_kwarg
        self.message = 'Passed "{version_kwarg}" argument value "{passed_version}" is invalid, \
            as model "{passed_name}" (passed as "{name_kwarg}" argument value) has no such version.'.format(
            passed_name=passed_name,
            passed_version=passed_version,
            name_kwarg=name_kwarg,
            version_kwarg=version_kwarg,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (
            InvalidModelVersion,
            (
                self.passed_name,
                self.passed_version,
                self.name_kwarg,
                self.version_kwarg,
            ),
        )


class KwargNotPassedWithUploadType(ServerException):
    """Define KwargNotPassedWithUploadType exception."""

    def __init__(self, kwarg, passed_upload_model_type):
        self.passed_upload_model_type = passed_upload_model_type
        self.kwarg = kwarg
        self.message = 'Argument "{kwarg}" cannot be ommitted with upload_model_type UploadModelType.{passed_upload_model_type}.'.format(
            kwarg=kwarg,
            passed_upload_model_type=passed_upload_model_type,
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (KwargNotPassedWithUploadType, (self.kwarg, self.passed_upload_model_type))


class ModelTypeIsNotFound(ServerException):
    """Define ModelTypeIsNotFound exception."""

    def __init__(self):
        self.message = "Model type is not found. You must inherit the desired template."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ModelTypeIsNotFound, ())


class ExistingModelWithOtherType(ServerException):
    """Define ExistingModelWithOtherType exception."""

    def __init__(self, name):
        self.name = name
        self.message = (
            f'The other model type with name "{name}" exists.'
            f"You cannot upload a model with the same name as an existing model of a different type."
        )

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (ExistingModelWithOtherType, (self.name,))


class InvalidExperimentName(ServerException):
    """Define InvalidExperimentName exception."""

    def __init__(self, model_type, exp_name):
        self.model_type = model_type
        self.exp_name = exp_name
        self.message = f"You can't specify '{exp_name}' experiment name for model type '{model_type}' upload."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidExperimentName, (self.model_type, self.exp_name))


class InvalidUploadModelMode(ServerException):
    """Define InvalidUploadModelMode exception."""

    def __init__(self, model_type, upload_mode):
        self.model_type = model_type
        self.upload_mode = upload_mode
        self.message = f"You can't specify upload mode '{upload_mode}' for '{model_type}' model."

        super().__init__(self.message)

    def __reduce__(self):
        """Define reduce method to make exception picklable."""
        return (InvalidUploadModelMode, (self.model_type, self.upload_mode))



"""Collector for fetching texts marked with ceritain topic, as well as negative examples."""
import os
from typing import List

import pandas as pd
from ML_management.collectors.collector_pattern import CollectorPattern
from ML_management.collectors.topic_markers.api_schema import Query
from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.operation import Operation


class TopicMarkersCollector(CollectorPattern):
    """Collector class for fetching texts marked with ceritain topic, as well as negative examples."""

    def __init__(self):
        # can not import GRAPHQL_URL from server's env_variables file
        self.graphql_url = os.environ.get("GRAPHQL_URL", "http://localhost:9000/graphql")
        self.endpoint = HTTPEndpoint(self.graphql_url)
        # TODO for now, recreate csv file every time at a hardcoded location

    @staticmethod
    def get_json_schema():
        """Return json schema."""
        schema = {
            "type": "object",
            "properties": {
                "local_path": {"type": "string"},
                "marker": {"type": "string"},
                "start_date": {"type": "integer"},
                "end_date": {"type": "integer"},
            },
            "required": [
                "marker",
            ],
            "additionalProperties": False,
        }

        return schema

    def __save_texts(self, texts: List[str], labels: List[int], path: str):
        """Save texts and labels to csv by passed path."""
        # path is supposed to be a full path to file
        data_dict = {
            "text": texts,
            "label": labels,
        }
        dataframe = pd.DataFrame(data_dict)
        # shuffle everything
        dataframe = dataframe.sample(frac=1).reset_index(drop=True)
        # this should work correctly with commas and quotes in text...
        dataframe.to_csv(path, index=False)

    def __gql_query(self, form: dict) -> dict:
        """Make GraphQL query to fetch new data from IE module."""
        op = Operation(Query)

        query = op.list_text_from_document_with_marker(form=form)
        query.marker_text()
        query.not_marker_text()

        data = self.endpoint(op)
        return data

    def set_data(self, *args, **kwargs):
        """Fetch texts with passed marker, serialize into given dir and return path."""
        local_path = kwargs.get("local_path", "topic_markers.csv")
        form = {
            "marker": kwargs["marker"],
        }
        if "start_date" in kwargs:
            form["startDate"] = kwargs["start_date"]
        if "end_date" in kwargs:
            form["endDate"] = kwargs["end_date"]

        raw_data = self.__gql_query(form)
        marker_texts = raw_data["data"]["listTextFromDocumentWithMarker"]["markerText"]
        not_marker_texts = raw_data["data"]["listTextFromDocumentWithMarker"]["notMarkerText"]
        # TODO do we raise if there are no examples whatsoever?
        all_labels = [1] * len(marker_texts) + [0] * len(not_marker_texts)
        all_texts = marker_texts + not_marker_texts
        self.__save_texts(texts=all_texts, labels=all_labels, path=local_path)

        return local_path



"""Functions for internal usage."""

from ML_management.mlmanagement import mlmanagement

import mlflow


def _load_model_type(run_id, unwrap: bool = True):
    """Load model from local path."""
    local_path = mlmanagement.MlflowClient().download_artifacts(run_id, "")
    loaded_model = mlflow.pyfunc.load_model(local_path)
    if unwrap:
        artifacts_dict = loaded_model._model_impl.context._artifacts
        loaded_model = loaded_model.unwrap_python_model()
        loaded_model.artifacts = artifacts_dict
    return loaded_model



"""Executor template for custom executor."""
from abc import ABC, abstractmethod
from typing import List, Optional

from ML_management.executor_template.upload_model_mode import UploadModelMode
from ML_management.mlmanagement import mlmanagement, variables
from ML_management.mlmanagement.module_finder import ModuleFinder
from ML_management.mlmanagement.variables import EXPERIMENT_NAME_FOR_EXECUTOR
from ML_management.models.model_type_to_methods_map import ModelMethodName
from mlflow.pyfunc import PythonModel


class JobExecutorPattern(PythonModel, ABC):
    """Define custom job executor."""

    def __init__(self, executor_name: str, desired_model_methods: List[ModelMethodName], executor_upload_model_mode: UploadModelMode):
        """
        Init JobExecutorPattern class.

        :param executor_name: The name of the executor
        :param desired_model_methods: Specify list of desired model methods for this executor
        :param executor_upload_model_mode: How to log model after job
        """
        self.executor_name = executor_name
        self.desired_model_methods = desired_model_methods
        self.executor_upload_model_mode = executor_upload_model_mode

        # That parameters will be set automatically while loading the model.
        """
        :param self.artifacts: A dictionary containing ``<name, local_path>`` entries.
        """
        self.artifacts: dict = {}

        # That parameters will be set automatically in job before the 'execute' func would be executed.
        """
        :param self.model: python model class
        :param self.dataset: object for dataset
        :param self.model_methods_parameters: the dict of parameters for each desired_model_methods.
            One could use it in execute() function like that:
                def execute(self):
                    self.model.train_function(**self.model_methods_parameters[ModelMethodName.train_function])
            In that case method 'execute' calls train_function method of the model with corresponding parameters for that method
            See examples in default_executors folder.
        :param self.model_methods_schema: the dict of schema parameters for each desired_model_methods.
        """
        self.model = None
        self.dataset = None
        self.model_methods_parameters: dict = {}
        self.model_methods_schema: dict = {}

    @abstractmethod
    def execute(self, **executor_params):
        """
        Do execution step.

        Parameter self.model with the desired model will be set automatically in the job before 'execute' execution.
        To get data_path use self.data_path parameter, which also will be set in the job.
        'executor_methods_params' are executor parameters. One has to define it as ordinary kwargs with type annotation.
        Also, you could use self.model_methods_parameters for call desired model method with right params.
        return param: artifacts: A dictionary containing ``<name, artifact_uri>`` entries.
                      For example, consider the following ``artifacts`` dictionary::

                        {
                            "my_file": "s3://my-bucket/path/to/my/file",
                            "my_file2": "/home/username/path/to/my/file"
                        }

                      In this case, the ``"my_file"`` artifact is downloaded from S3.
                      The ``"my_file2"`` artifact is downloaded from local path.

                      If ``None``, no artifacts are added to the model.
        """
        raise NotImplementedError

    def upload_executor(
        self,
        pip_requirements=None,
        extra_pip_requirements=None,
        conda_env=None,
        artifacts: Optional[dict] = None,
        extra_modules_names: Optional[List[str]] = None,
        used_modules_names: Optional[List[str]] = None,
    ):
        """
        Upload wrapper to MLmanagement server.

        :param pip_requirements: {{ pip_requirements }}

        :param extra_pip_requirements: {{ extra_pip_requirements }}
        `pip_requirements` and 'extra_pip_requirements' must be either a string path to a pip requirements file on the
            local filesystem or an iterable of pip requirement strings.

        :param conda_env: {{ conda_env }}
        'conda_env' must be a dict specifying the conda environment for this model.

        :param artifacts: A dictionary containing ``<name, artifact_uri>`` entries. Remote artifact URIs
                          are resolved to absolute filesystem paths, producing a dictionary of
                          ``<name, absolute_path>`` entries. ``python_model`` can reference these
                          resolved entries as the ``artifacts`` property of the ``context`` parameter
                          in :func:`PythonModel.load_context() <mlflow.pyfunc.PythonModel.load_context>`
                          and :func:`PythonModel.predict() <mlflow.pyfunc.PythonModel.predict>`.
                          For example, consider the following ``artifacts`` dictionary::

                            {
                                "my_file": "s3://my-bucket/path/to/my/file"
                            }

                          In this case, the ``"my_file"`` artifact is downloaded from S3. The
                          ``python_model`` can then refer to ``"my_file"`` as an absolute filesystem
                          path via ``context.artifacts["my_file"]``.

                          If ``None``, no artifacts are added to the executor.

        :param extra_modules_names: names of modules that should be pickled by value
            in addition to auto-detected modules.

        :param used_modules_names: modules that should be pickled by value, disables the auto-detection of modules.
        """
        old_experiment_name = variables.active_experiment_name
        mlmanagement.set_experiment(EXPERIMENT_NAME_FOR_EXECUTOR)
        try:
            with mlmanagement.start_run(nested=True):
                mlmanagement.log_model(
                    artifact_path="",
                    artifacts=artifacts,
                    python_model=self,
                    registered_model_name=self.executor_name,
                    pip_requirements=pip_requirements,
                    extra_pip_requirements=extra_pip_requirements,
                    conda_env=conda_env,
                    extra_modules_names=extra_modules_names,
                    used_modules_names=used_modules_names,
                    root_module_name=ModuleFinder.get_my_caller_module_name(),
                )
        except Exception as err:
            raise err
        finally:
            variables.active_experiment_name = old_experiment_name
